﻿using System.Windows;
using System.Windows.Controls;
using System.Diagnostics;
using System.Windows.Data;

namespace Matarillo
{
    /// <summary>
    /// Silverlight ToolkitのHeaderedItemsControlを借用。
    /// </summary>
    public class HeaderedItemsControl : ItemsControl, IContainer
    {
        /// <summary>
        /// HeaderプロパティにItemsControlのアイテムが設定されているかどうかを示すプロパティ。
        /// </summary>
        internal bool HeaderIsItem { get; set; }

        #region Headerプロパティ
        /// <summary>
        /// HeaderedItemsControlのラベルを表すプロパティ。
        /// デフォルト値はnull。
        /// ヘッダーはstringやUIElementやその他の型でありえる。
        /// </summary>
        public object Header
        {
            get { return GetValue(HeaderProperty); }
            set { SetValue(HeaderProperty, value); }
        }

        /// <summary>
        /// Header依存プロパティを表す。
        /// </summary>
        /// <remarks>
        /// Note: WPFでは、HeaderedContentControlのHeaderPropertyに対してAddOwnerを呼び出して定義する。
        /// </remarks>
        public static readonly DependencyProperty HeaderProperty =
            DependencyProperty.Register(
                "Header",
                typeof(object),
                typeof(HeaderedItemsControl),
                new PropertyMetadata(OnHeaderPropertyChanged));

        /// <summary>
        /// Headerプロパティが更新されたときのハンドラ。
        /// </summary>
        /// <param name="d">
        /// Headerが変更されたHeaderedItemsControl。
        /// </param>
        /// <param name="e">イベント引数。</param>
        private static void OnHeaderPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            HeaderedItemsControl source = d as HeaderedItemsControl;
            source.OnHeaderChanged(e.OldValue, e.NewValue);
        }
        #endregion

        #region HeaderTemplateプロパティ
        /// <summary>
        /// コントロールのヘッダーを表示する際のデータテンプレートを表すプロパティ。
        /// デフォルト値はnull。
        /// </summary>
        public DataTemplate HeaderTemplate
        {
            get { return GetValue(HeaderTemplateProperty) as DataTemplate; }
            set { SetValue(HeaderTemplateProperty, value); }
        }

        /// <summary>
        /// HeaderTemplate依存プロパティを表す。
        /// </summary>
        /// <remarks>
        /// Note: WPFでは、HeaderedContentControlのHeaderTemplatePropertyに対してAddOwnerを呼び出して定義する。
        /// </remarks>
        public static readonly DependencyProperty HeaderTemplateProperty =
            DependencyProperty.Register(
                "HeaderTemplate",
                typeof(DataTemplate),
                typeof(HeaderedItemsControl),
                new PropertyMetadata(OnHeaderTemplatePropertyChanged));

        /// <summary>
        /// HeaderTemplateプロパティが更新されたときのハンドラ。
        /// </summary>
        /// <param name="d">
        /// HeaderTemplateが変更されたHeaderedItemsControl。
        /// </param>
        /// <param name="e">Event arguments.</param>
        private static void OnHeaderTemplatePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            HeaderedItemsControl source = d as HeaderedItemsControl;
            DataTemplate oldHeaderTemplate = e.OldValue as DataTemplate;
            DataTemplate newHeaderTemplate = e.NewValue as DataTemplate;
            source.OnHeaderTemplateChanged(oldHeaderTemplate, newHeaderTemplate);
        }
        #endregion

        #region ItemContainerStyleプロパティ
        /// <summary>
        /// アイテムを格納するコンテナ要素に適用するスタイル表すプロパティ。
        /// </summary>
        public Style ItemContainerStyle
        {
            get { return GetValue(ItemContainerStyleProperty) as Style; }
            set { SetValue(ItemContainerStyleProperty, value); }
        }

        /// <summary>
        /// ItemContainerStyle依存プロパティを表す。
        /// </summary>
        public static readonly DependencyProperty ItemContainerStyleProperty =
            DependencyProperty.Register(
                "ItemContainerStyle",
                typeof(Style),
                typeof(HeaderedItemsControl),
                new PropertyMetadata(null, OnItemContainerStylePropertyChanged));

        /// <summary>
        /// ItemContainerStyleプロパティが更新されたときのハンドラ。
        /// </summary>
        /// <param name="d">
        /// ItemContainerStyleが変更されたHeaderedItemsControl。
        /// </param>
        /// <param name="e">Event arguments.</param>
        private static void OnItemContainerStylePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            HeaderedItemsControl source = d as HeaderedItemsControl;
            Style value = e.NewValue as Style;
            source.ItemContainerGenerator.UpdateItemContainerStyle(value);
        }
        #endregion

        /// <summary>
        /// このコントロールに関連付けられたItemContainerGeneratorを取得。
        /// </summary>
        public ItemContainerGenerator ItemContainerGenerator { get; private set; }

        /// <summary>
        /// HeaderedItemsControlの新しいインスタンスを初期化。
        /// </summary>
        public HeaderedItemsControl()
        {
            DefaultStyleKey = typeof(HeaderedItemsControl);
            ItemContainerGenerator = new ItemContainerGenerator(this);
        }

        /// <summary>
        /// Headerプロパティが変更されたときに呼び出される。
        /// </summary>
        /// <param name="oldHeader">
        /// Headerプロパティの以前の値。
        /// </param>
        /// <param name="newHeader">
        /// Headerプロパティの新しい値。
        /// </param>
        protected virtual void OnHeaderChanged(object oldHeader, object newHeader)
        {
        }

        /// <summary>
        /// HeaderTemplateプロパティが変更されたときに呼び出される。
        /// </summary>
        /// <param name="oldHeaderTemplate">
        /// HeaderTemplateプロパティの以前の値。
        /// </param>
        /// <param name="newHeaderTemplate">
        /// HeaderTemplateプロパティの新しい値。
        /// </param>
        protected virtual void OnHeaderTemplateChanged(DataTemplate oldHeaderTemplate, DataTemplate newHeaderTemplate)
        {
        }

        /// <summary>
        /// コントロールにテンプレートを適用する。
        /// </summary>
        public override void OnApplyTemplate()
        {
            ItemContainerGenerator.OnApplyTemplate();
            base.OnApplyTemplate();
        }

        /// <summary>
        /// 指定したコンテナが指定したアイテムを表示する準備をする。
        /// </summary>
        /// <param name="element">
        /// 指定したアイテムを表示するのに使うコンテナ。
        /// </param>
        /// <param name="item">表示するアイテム。</param>
        protected override void PrepareContainerForItemOverride(DependencyObject element, object item)
        {
            ItemContainerGenerator.PrepareContainerForItemOverride(element, item, ItemContainerStyle);
            base.PrepareContainerForItemOverride(element, item);
        }

        /// <summary>
        /// PrepareContainerForItemOverrideの効果を取り消す。
        /// </summary>
        /// <param name="element">コンテナ。</param>
        /// <param name="item">アイテム。</param>
        protected override void ClearContainerForItemOverride(DependencyObject element, object item)
        {
            ItemContainerGenerator.ClearContainerForItemOverride(element, item);
            base.ClearContainerForItemOverride(element, item);
        }

        /// <summary>
        /// 準備をする。
        /// </summary>
        /// <param name="item">コンテナに含めるアイテム。</param>
        /// <param name="parentItemsControl">親のItemsControl。</param>
        /// <param name="parentItemContainerStyle">親のItemsControlに適用するItemContainerStyle。</param>
        public void Prepare(object item, ItemsControl parentItemsControl, Style parentItemContainerStyle)
        {
            if (this == item) return;
            // ItemsControlのプロパティを子にコピーする
            DataTemplate parentItemTemplate = parentItemsControl.ItemTemplate;
            if (parentItemTemplate != null)
            {
                this.SetValue(HeaderedItemsControl.ItemTemplateProperty, parentItemTemplate);
            }
            if (parentItemContainerStyle != null && HasDefaultValue(HeaderedItemsControl.ItemContainerStyleProperty))
            {
                this.SetValue(HeaderedItemsControl.ItemContainerStyleProperty, parentItemContainerStyle);
            }

            // 親のHeaderプロパティを子にコピーする
            if (this.HeaderIsItem || HasDefaultValue(HeaderedItemsControl.HeaderProperty))
            {
                this.Header = item;
                this.HeaderIsItem = true;
            }
            if (parentItemTemplate != null)
            {
                this.SetValue(HeaderedItemsControl.HeaderTemplateProperty, parentItemTemplate);
            }
            if (parentItemContainerStyle != null && this.Style == null)
            {
                this.SetValue(HeaderedItemsControl.StyleProperty, parentItemContainerStyle);
            }

            // 階層化テンプレートの準備をする
            HierarchicalDataTemplate headerTemplate = parentItemTemplate as HierarchicalDataTemplate;
            if (headerTemplate != null)
            {
                if (headerTemplate.ItemsSource != null && HasDefaultValue(HeaderedItemsControl.ItemsSourceProperty))
                {
                    this.SetBinding(
                        HeaderedItemsControl.ItemsSourceProperty,
                        new Binding
                        {
                            Converter = headerTemplate.ItemsSource.Converter,
                            ConverterCulture = headerTemplate.ItemsSource.ConverterCulture,
                            ConverterParameter = headerTemplate.ItemsSource.ConverterParameter,
                            Mode = headerTemplate.ItemsSource.Mode,
                            NotifyOnValidationError = headerTemplate.ItemsSource.NotifyOnValidationError,
                            Path = headerTemplate.ItemsSource.Path,
                            Source = this.Header,
                            ValidatesOnExceptions = headerTemplate.ItemsSource.ValidatesOnExceptions
                        });
                }
                if (headerTemplate.IsItemTemplateSet && this.ItemTemplate == parentItemTemplate)
                {
                    this.ClearValue(HeaderedItemsControl.ItemTemplateProperty);
                    if (headerTemplate.ItemTemplate != null)
                    {
                        this.ItemTemplate = headerTemplate.ItemTemplate;
                    }
                }
                if (headerTemplate.IsItemContainerStyleSet && this.ItemContainerStyle == parentItemContainerStyle)
                {
                    this.ClearValue(HeaderedItemsControl.ItemContainerStyleProperty);
                    if (headerTemplate.ItemContainerStyle != null)
                    {
                        this.ItemContainerStyle = headerTemplate.ItemContainerStyle;
                    }
                }
            }
        }

        /// <summary>
        /// プロパティがデフォルト値を持つか調べる。
        /// </summary>
        /// <param name="property">調査するプロパティ。</param>
        /// <returns>
        /// プロパティがデフォルト値を持つときTrue。
        /// </returns>
        private bool HasDefaultValue(DependencyProperty property)
        {
            return this.ReadLocalValue(property) == DependencyProperty.UnsetValue;
        }
    }
}
