﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Media;

namespace Matarillo
{
    /// <summary>
    /// Silverlight ToolkitのItemContainerGeneratorを借用。
    /// </summary>
    public class ItemContainerGenerator
    {
        private static int count = 0;
        private int number = 0;

        private ItemsControl ItemsControl { get; set; }

        private IDictionary<DependencyObject, object> ContainersToItems { get; set; }

        /// <summary>
        /// ItemsHostプロパティの値を保持しておくフィールド。
        /// </summary>
        private Panel _itemsHost;

        /// <summary>
        /// ItemsControlの親パネルを取得する。
        /// ItemsControlがヴィジュアルツリー上に存在し、
        /// 子アイテムを保持するコンテナを生成した後で有効。
        /// </summary>
        internal Panel ItemsHost
        {
            get
            {
                // Lookup the ItemsHost if we haven't already cached it.
                if (_itemsHost == null)
                {
                    // Get any live container
                    if (ContainersToItems.Count <= 0)
                    {
                        return null;
                    }
                    DependencyObject container = ContainersToItems.First().Key;

                    // Get the parent of the container
                    _itemsHost = VisualTreeHelper.GetParent(container) as Panel;
                }

                return _itemsHost;
            }
        }

        /// <summary>
        /// ItemContainerGeneratorの新インスタンスを生成する。
        /// </summary>
        /// <param name="control">
        /// ItemContainerGeneratorから追跡されるItemsControl。
        /// </param>
        internal ItemContainerGenerator(ItemsControl control)
        {
            number = count++;
            Debug.WriteLine("ItemContainerGenerator{0} created. ({1})", number, control.GetType().Name);
            Debug.Assert(control != null, "control cannot be null!");
            ItemsControl = control;
            ContainersToItems = new Dictionary<DependencyObject, object>();
        }

        /// <summary>
        /// テンプレートをItemsControlに適用する。
        /// </summary>
        internal void OnApplyTemplate()
        {
            Debug.WriteLine("ItemContainerGenerator{0}.OnApplyTemplate", number);
            _itemsHost = null;
            ContainersToItems.Clear();
        }

        /// <summary>
        /// 指定したコンテナが指定したアイテムを表示できるよう準備する。
        /// </summary>
        /// <param name="element">
        /// 指定したアイテムを表示するのに使うコンテナ。
        /// </param>
        /// <param name="item">表示するアイテム。</param>
        /// <param name="parentItemContainerStyle">
        /// 親のItemsControlに適用するItemContainerStyle。
        /// </param>
        internal void PrepareContainerForItemOverride(DependencyObject element, object item, Style parentItemContainerStyle)
        {
            Debug.WriteLine("ItemContainerGenerator{0}.PrepareContainerForItemOverride", number);
            // コンテナをアイテムと関連付ける
            ContainersToItems[element] = item;

            // ItemContainerStyleをアイテムに適用する
            Control control = element as Control;
            if (parentItemContainerStyle != null && control != null && control.Style == null)
            {
                control.SetValue(Control.StyleProperty, parentItemContainerStyle);
            }

            // コンテナの準備をする
            IContainer container = element as IContainer;
            if (container != null)
            {
                container.Prepare(item, ItemsControl, parentItemContainerStyle);
            }
        }

        /// <summary>
        /// PrepareContainerForItemOverrideの効果を取り消す。
        /// </summary>
        /// <param name="element">コンテナ。</param>
        /// <param name="item">コンテナに含まれるアイテム。</param>
        [SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "item", Justification = "Matching ItemsControl signature.")]
        internal void ClearContainerForItemOverride(DependencyObject element, object item)
        {
            ContainersToItems.Remove(element);
        }

        /// <summary>
        /// ItemContainerStyleが変更されたときに、生成された子アイテムのスタイルを更新する。
        /// </summary>
        /// <param name="itemContainerStyle">ItemContainerStyleオブジェクト。</param>
        /// <remarks>
        /// SilverlightではStyleを複数回設定することができないので、
        /// ここでは、まだスタイルが設定されていないアイテムへの適用を試みるだけである。
        /// </remarks>
        internal void UpdateItemContainerStyle(Style itemContainerStyle)
        {
            if (itemContainerStyle == null)
            {
                return;
            }

            Panel itemsHost = ItemsHost;
            if (itemsHost == null || itemsHost.Children == null)
            {
                return;
            }

            foreach (UIElement element in itemsHost.Children)
            {
                FrameworkElement obj = element as FrameworkElement;
                if (obj.Style == null)
                {
                    obj.Style = itemContainerStyle;
                }
            }
        }

        /// <summary>
        /// ItemCollectionのインデックスに対応するアイテムのコンテナを返す。
        /// </summary>
        /// <param name="index">インデックス。</param>
        /// <returns>
        /// 対応するアイテムのコンテナ。アイテムが生成されていないときはnull。
        /// </returns>
        public DependencyObject ContainerFromIndex(int index)
        {
            Panel host = ItemsHost;
            if (host == null || host.Children == null || index < 0 || index >= host.Children.Count)
            {
                return null;
            }

            return host.Children[index];
        }

        /// <summary>
        /// 指定されたアイテムのコンテナを返す。
        /// </summary>
        /// <param name="item">コンテナを探したいアイテム。</param>
        /// <returns>
        /// 対応するアイテムのコンテナ。
        /// アイテムがItemCollectionに含まれていないか、コンテナが生成されていないときはnull。
        /// </returns>
        public DependencyObject ContainerFromItem(object item)
        {
            // TODO: Consider replacing dictionary with a list so that search
            // will be consistent regardless of dictionary entries
            foreach (KeyValuePair<DependencyObject, object> mapping in ContainersToItems)
            {
                if (object.Equals(item, mapping.Value))
                {
                    return mapping.Key;
                }
            }
            return null;
        }

        /// <summary>
        /// 指定されたコンテナのインデックスを返す。
        /// </summary>
        /// <param name="container">
        /// アイテムに対応するコンテナ。
        /// </param>
        /// <returns>
        /// 指定されたコンテナのインデックス。
        /// </returns>
        public int IndexFromContainer(DependencyObject container)
        {
            if (container == null)
            {
                throw new ArgumentNullException("container");
            }

            UIElement element = container as UIElement;
            if (element == null)
            {
                return -1;
            }

            Panel host = ItemsHost;
            if (host == null || host.Children == null)
            {
                return -1;
            }

            return host.Children.IndexOf(element);
        }

        /// <summary>
        /// 指定されたコンテナが含むアイテムを返す。
        /// </summary>
        /// <param name="container">
        /// アイテムに対応するコンテナ。
        /// </param>
        /// <returns>
        /// コンテナがアイテムを含む場合はそのアイテム、含んでいない場合はコンテナ自身。
        /// </returns>
        public object ItemFromContainer(DependencyObject container)
        {
            if (container == null)
            {
                throw new ArgumentNullException("container");
            }

            object item = container;
            ContainersToItems.TryGetValue(container, out item);
            return item;
        }
    }
}
