﻿using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace Matarillo
{
    public class TreeControlBase<T> : ItemsControl where T : Control, new()
    {
        private IDictionary<DependencyObject, object> containers;

        private Panel parentPanel;

        private Panel ParentPanelCache
        {
            get
            {
                if (parentPanel == null)
                {
                    if (containers.Count <= 0)
                    {
                        return null;
                    }
                    DependencyObject container = containers.First().Key;
                    parentPanel = VisualTreeHelper.GetParent(container) as Panel;
                }
                return parentPanel;
            }
        }

        #region ItemContainerStyleプロパティ
        /// <summary>
        /// ItemContainerStyleプロパティ。
        /// </summary>
        public Style ItemContainerStyle
        {
            get { return GetValue(ItemContainerStyleProperty) as Style; }
            set { SetValue(ItemContainerStyleProperty, value); }
        }

        /// <summary>
        /// ItemContainerStyle依存プロパティ。
        /// </summary>
        public static readonly DependencyProperty ItemContainerStyleProperty =
            DependencyProperty.Register(
                "ItemContainerStyle",
                typeof(Style),
                typeof(TreeControlBase<T>),
                new PropertyMetadata(null, OnItemContainerStylePropertyChanged));

        /// <summary>
        /// ItemContainerStyleプロパティが更新されたときのハンドラ。
        /// </summary>
        /// <param name="d">
        /// ItemContainerStyleが変更されたTreeControlBase。
        /// </param>
        /// <param name="e">イベント引数。</param>
        private static void OnItemContainerStylePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TreeControlBase<T> self = d as TreeControlBase<T>;
            Style value = e.NewValue as Style;
            self.UpdateItemContainerStyle(value);
        }
        #endregion

        public TreeControlBase()
        {
            containers = new Dictionary<DependencyObject, object>();
        }

        /// <summary>
        /// コントロールにテンプレートを適用する。
        /// </summary>
        public override void OnApplyTemplate()
        {
            parentPanel = null;
            containers.Clear();
            base.OnApplyTemplate();
        }

        protected override DependencyObject GetContainerForItemOverride()
        {
            return new T();
        }

        protected override bool IsItemItsOwnContainerOverride(object item)
        {
            return item is T;
        }

        protected override void PrepareContainerForItemOverride(DependencyObject element, object item)
        {
            containers[element] = item;
            Control control = element as Control;
            if (this.ItemContainerStyle != null && control != null && control.Style == null)
            {
                control.SetValue(Control.StyleProperty, this.ItemContainerStyle);
            }
            TreeControlBase<T> node = element as TreeControlBase<T>;
            if (node != null)
            {
                node.Prepare(item, this);
            }
            base.PrepareContainerForItemOverride(element, item);
        }

        protected override void ClearContainerForItemOverride(DependencyObject element, object item)
        {
            containers.Remove(element);
            base.ClearContainerForItemOverride(element, item);
        }

        private void UpdateItemContainerStyle(Style itemContainerStyle)
        {
            if (itemContainerStyle == null)
            {
                return;
            }

            Panel parentPanel = this.ParentPanelCache;
            if (parentPanel == null || parentPanel.Children == null)
            {
                return;
            }

            foreach (UIElement element in parentPanel.Children)
            {
                FrameworkElement obj = element as FrameworkElement;
                if (obj.Style == null)
                {
                    obj.Style = itemContainerStyle;
                }
            }
        }

        protected virtual void Prepare(object item, TreeControlBase<T> parent)
        {
        }
    }
}
