﻿using System.Collections.Specialized;
using System.Windows;
using System.Windows.Data;

namespace Matarillo
{
    [StyleTypedProperty(Property = "ItemContainerStyle", StyleTargetType = typeof(TreeGridItem))]
    public class TreeGridItem : TreeControlBase<TreeGridItem>
    {
        private bool HeaderIsItem { get; set; }

        #region Headerプロパティ
        /// <summary>
        /// HeaderedItemsControlのラベルを表すプロパティ。
        /// デフォルト値はnull。
        /// ヘッダーはstringやUIElementやその他の型でありえる。
        /// </summary>
        public object Header
        {
            get { return GetValue(HeaderProperty); }
            set { SetValue(HeaderProperty, value); }
        }

        /// <summary>
        /// Header依存プロパティ。
        /// </summary>
        public static readonly DependencyProperty HeaderProperty =
            DependencyProperty.Register(
                "Header",
                typeof(object),
                typeof(TreeGridItem),
                new PropertyMetadata(OnHeaderPropertyChanged));

        private static void OnHeaderPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TreeGridItem self = d as TreeGridItem;
            self.OnHeaderChanged(e.OldValue, e.NewValue);
        }

        protected virtual void OnHeaderChanged(object oldHeader, object newHeader)
        {
        }
        #endregion

        #region HeaderTemplateプロパティ
        /// <summary>
        /// コントロールのヘッダーを表示する際のデータテンプレートを表すプロパティ。
        /// デフォルト値はnull。
        /// </summary>
        public DataTemplate HeaderTemplate
        {
            get { return GetValue(HeaderTemplateProperty) as DataTemplate; }
            set { SetValue(HeaderTemplateProperty, value); }
        }

        /// <summary>
        /// HeaderTemplate依存プロパティ。
        /// </summary>
        public static readonly DependencyProperty HeaderTemplateProperty =
            DependencyProperty.Register(
                "HeaderTemplate",
                typeof(DataTemplate),
                typeof(TreeGridItem),
                new PropertyMetadata(OnHeaderTemplatePropertyChanged));

        private static void OnHeaderTemplatePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            TreeGridItem self = d as TreeGridItem;
            DataTemplate oldHeaderTemplate = e.OldValue as DataTemplate;
            DataTemplate newHeaderTemplate = e.NewValue as DataTemplate;
            self.OnHeaderTemplateChanged(oldHeaderTemplate, newHeaderTemplate);
        }

        protected virtual void OnHeaderTemplateChanged(DataTemplate oldHeaderTemplate, DataTemplate newHeaderTemplate)
        {
        }
        #endregion

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        public TreeGridItem()
        {
            DefaultStyleKey = typeof(TreeGridItem);
        }

        protected override void Prepare(object item, TreeControlBase<TreeGridItem> parent)
        {
            if (this == item) return;
            DataTemplate parentItemTemplate = parent.ItemTemplate;
            Style parentItemContainerStyle = parent.ItemContainerStyle;

            // ItemTemplateプロパティを親からコピーする
            if (parentItemTemplate != null)
            {
                this.SetValue(TreeGridItem.ItemTemplateProperty, parentItemTemplate);
            }
            if (parentItemContainerStyle != null && HasDefaultValue(TreeGridItem.ItemContainerStyleProperty))
            {
                this.SetValue(TreeGridItem.ItemContainerStyleProperty, parentItemContainerStyle);
            }

            // Headerプロパティにアイテムを設定する
            if (this.HeaderIsItem || HasDefaultValue(TreeGridItem.HeaderProperty))
            {
                this.Header = item;
                this.HeaderIsItem = true;
            }
            // HeaderTemplateプロパティを親のItemTemplateからコピーする
            if (parentItemTemplate != null)
            {
                this.SetValue(TreeGridItem.HeaderTemplateProperty, parentItemTemplate);
            }
            // Styleプロパティを親からコピーする
            if (parentItemContainerStyle != null && this.Style == null)
            {
                this.SetValue(TreeGridItem.StyleProperty, parentItemContainerStyle);
            }

            // 階層化テンプレートを設定する
            HierarchicalDataTemplate headerTemplate = parentItemTemplate as HierarchicalDataTemplate;
            if (headerTemplate != null)
            {
                if (headerTemplate.ItemsSource != null && HasDefaultValue(TreeGridItem.ItemsSourceProperty))
                {
                    this.SetBinding(
                        TreeGridItem.ItemsSourceProperty,
                        new Binding
                        {
                            Converter = headerTemplate.ItemsSource.Converter,
                            ConverterCulture = headerTemplate.ItemsSource.ConverterCulture,
                            ConverterParameter = headerTemplate.ItemsSource.ConverterParameter,
                            Mode = headerTemplate.ItemsSource.Mode,
                            NotifyOnValidationError = headerTemplate.ItemsSource.NotifyOnValidationError,
                            Path = headerTemplate.ItemsSource.Path,
                            Source = this.Header,
                            ValidatesOnExceptions = headerTemplate.ItemsSource.ValidatesOnExceptions
                        });
                }
                if (headerTemplate.IsItemTemplateSet && this.ItemTemplate == parentItemTemplate)
                {
                    this.ClearValue(TreeGridItem.ItemTemplateProperty);
                    if (headerTemplate.ItemTemplate != null)
                    {
                        this.ItemTemplate = headerTemplate.ItemTemplate;
                    }
                }
                if (headerTemplate.IsItemContainerStyleSet && this.ItemContainerStyle == parentItemContainerStyle)
                {
                    this.ClearValue(TreeGridItem.ItemContainerStyleProperty);
                    if (headerTemplate.ItemContainerStyle != null)
                    {
                        this.ItemContainerStyle = headerTemplate.ItemContainerStyle;
                    }
                }
            }
        }

        /// <summary>
        /// プロパティがデフォルト値を持つか調べる。
        /// </summary>
        /// <param name="property">調査するプロパティ。</param>
        /// <returns>
        /// プロパティがデフォルト値を持つときTrue。
        /// </returns>
        private bool HasDefaultValue(DependencyProperty property)
        {
            return this.ReadLocalValue(property) == DependencyProperty.UnsetValue;
        }

    }
}
